/*******************************************************************************
 *                                                                             *
 *        Program : Exclusive Prefix Sum for vector                            *
 *                                                                             *
 *        Author  : Kim, Dongmin (kimd@top.cis.syr.edu)                        *
 *                                                                             *
 *        This program simulates the prefix sum algorithm for vector.          *
 *                                                                             *
 *        The vector size should be multiple of total processor number.        *
 *                                                                             *
 *******************************************************************************/


#include <cm/cmmd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define Max   1024
#define fail  0
#define null  0
#define true  1
#define false 0

#define msg_size      4        /* integer */

/*******************************************************************************
 *                                                                             *
 *   Global variables                                                          *
 *                                                                             *
 *******************************************************************************/
int    my_id;  /* The ID of the Processor */
int    nproc;  /* Total # of Processors */
int    v_size; /* Vector Size */
int    digit;  /* binary digit */
int    Vector[Max];    /* original data */
int    Original[Max];  /* result data */
int    Sum[Max];     /* Sum from parent process */

int    partner;    /* The processor ID with which the data is exchanged */
int    right;      /* The processor ID of right child */
int    parent;     /* The processor ID of parent */

/** For Active Message Passing **/
int    rport = 8;
int    prs_dim_ack;
/********************************/

/** For Active Message Passing **/
/*******************************************************************************
 *   Procedure : prs_dim_handler                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
void prs_dim_handler() {

  prs_dim_ack = true;

}
/********************************/


show(id, start, size, vector)
     int id, start, size, *vector;
{
  int pn;
  int i;


  CMMD_fset_io_mode(stdout, CMMD_independent);
  for ( pn = 0 ; pn < nproc ; pn++ ) {
    if ( id == pn ) {
      printf("***********************************************\n");
      printf("Processor %2d => \n", id);
      for ( i = start ; i < start+size ; i++ )
	printf("%d  :  %3d\n", i, vector[i]);

      printf("***********************************************\n");
      fflush(stdout);
    }

    CMMD_sync_with_nodes();
  }
}


/*******************************************************************************
 *   Procedure : add                                                           *
 *   Input     : int start_index, int data_size, int *v_1, int *v_2            *
 *   Output    : None                                                          *
 *   Function  : Add the two arrays from the start_index by data_size.         *
 *******************************************************************************/
add(start_index, data_size, v_1, v_2)
     int start_index,
         data_size,
         *v_1, *v_2;
{
  int i;

  for ( i = start_index ; i < start_index + data_size ; i++ )
    v_1[i] += v_2[i];
}

/*******************************************************************************
 *   Procedure : add_n_store                                                   *
 *   Input     : int start_index, int data_size, int step, int *v_1, int *v_2  *
 *               int *v_3[]                                                    *
 *   Output    : None                                                          *
 *   Function  : Add the two arrays from the start_index by data_size, and     *
 *                store the right child's data.                                *
 *******************************************************************************/
add_n_store(start_index, data_size, step, v_1, v_2, v_3)
     int start_index,
         data_size,
         step,
         *v_1, *v_2, v_3[Max][Max];
{
  int i;

  if ( right == partner ) {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      v_1[i] += v_2[i];
      v_3[step][i] = v_2[i];
    }
  }
  else {
    for ( i = start_index ; i < start_index + data_size ; i++ ) {
      v_3[step][i] = v_1[i];
      v_1[i] += v_2[i];
    }
  }
}

/*******************************************************************************
 *   Procedure : Ilog2()                                                       *
 *   Input     : int n                                                         *
 *   Output    : log2(n)                                                       *
 *   Function  : return the integer value of log2(n) if it is integer.         *
 *               If not, return -1.                                            *
 *******************************************************************************/
int Ilog2(n)
     int n;
{
  int i = 0;
  int n1 = n;

  if ( n < 1 ) return -1;

  while ( n1 >>= 1 )
    ++i;

  if ( 1 << i != n )
    return -1;
  else
    return i;
}

/*******************************************************************************
 *   Procedure : pow2()                                                        *
 *   Input     : int x                                                         *
 *   Output    : 2^x                                                           *
 *   Function  : power of 2 for the input value                                *
 *******************************************************************************/
int pow2(x)
     int x;
{
  int p = 1;
  int i;

  for ( i = 0 ; i < x ; i++ )
    p <<= 1;

  return p;
}

/*******************************************************************************
 *   Procedure :                                                               *
 *   Input     :                                                               *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
ex_prefix(vector_size, input_v, p_result)
     int vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
	 *input_v,     /* input vector */
         *p_result;    /* exclusive prefix */
{
  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    Right[Max][Max];   /* Sums from right child process by step.
			       1st dimension is for step, and
			       2nd dimension is for the sums
			       from the right child process.          */
  int    in_mesg[Max];      /* received data */

  for ( i = 0 ; i < vector_size ; i++ )
    p_result[i] = input_v[i];

  size = sizeof(int);
  data_size = vector_size;

/* ===============
     PREFIX SUM
   =============== */

  start_recv = 0;
  digit = Ilog2(nproc);

/* Sum */

  for ( step = 0 ; step < digit ; step++ ) {
    data_size = data_size / 2;
    start_send = start_recv;
    check = my_id & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    partner = my_id ^ pow2(step);
    right = my_id | pow2(step);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &in_mesg[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner,
                rport,
                0,
                0,
                &p_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    add_n_store(start_recv, data_size, step, p_result, in_mesg, Right);
  }

/* Distribution */

  start_send = start_recv;
  for ( step = digit - 1 ; step >= 0 ; step-- ) {
    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    partner = my_id ^ pow2(step);
    right = my_id | pow2(step);

    if ( partner == right ) {

      /** For Active Message Passing **/

      prs_dim_ack = false;

      /* receive */
      CMAML_set_rport_addr         (rport, &p_result[start_recv]);
      CMAML_set_rport_byte_counter (rport, data_size*size);
      CMAML_set_rport_total_nbytes (rport, data_size*size);

      CMMD_sync_with_nodes();

      /* send */
      CMAML_scopy(partner,
		  rport,
		  0,
		  0,
		  &p_result[start_send],
		  data_size*size,
		  null,
		  null);  

      CMAML_poll_while_lt(&prs_dim_ack, true); 

      /********************************/

      for ( i = start_send ; i < start_send + data_size ; i++ )
	p_result[i] -= Right[step][i];
    }

    else { /* partner != right */
      for ( i = start_send ; i < start_send + data_size ; i++ )
	in_mesg[i] = p_result[i] - Right[step][i];

      /** For Active Message Passing **/

      prs_dim_ack = false;

      /* receive */
      CMAML_set_rport_addr         (rport, &p_result[start_recv]);
      CMAML_set_rport_byte_counter (rport, data_size*size);
      CMAML_set_rport_total_nbytes (rport, data_size*size);

      CMMD_sync_with_nodes();

      /* send */
      CMAML_scopy(partner,
		  rport,
		  0,
		  0,
		  &in_mesg[start_send],
		  data_size*size,
		  null,
		  null);  

      CMAML_poll_while_lt(&prs_dim_ack, true); 

      /********************************/
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

  /* for exclusive prefix */
  for ( i = 0 ; i < vector_size ; i++ )
    p_result[i] -= input_v[i];

  show(my_id, 0, vector_size, p_result);
}

/*******************************************************************************
 *   Procedure : reduction()                                                   *
 *   Input     : vector_size, p_result                                         *
 *   Output    :                                                               *
 *   Function  :                                                               *
 *******************************************************************************/
reduction(vector_size, input_v, r_result)
     int vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
	 *input_v,     /* input vector */
         *r_result;    /* exclusive prefix */
{
  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    in_mesg[Max];      /* received data */

/* ===============
     REDUCTION
   =============== */

  for ( i = 0 ; i < vector_size ; i++ )
    r_result[i] = input_v[i];

  start_recv = 0;
  digit = Ilog2(nproc);

  size = sizeof(int);
  data_size = vector_size;

/* Sum */

  for ( step = 0 ; step < digit ; step++ ) {
    data_size = data_size / 2;
    start_send = start_recv;
    check = my_id & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    partner = my_id ^ pow2(step);
    right = my_id | pow2(step);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &in_mesg[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner,
                rport,
                0,
                0,
                &r_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    add(start_recv, data_size, r_result, in_mesg);
  }

/* all-to-all BC */

  start_send = start_recv;
  for ( step = digit - 1 ; step >= 0 ; step-- ) {
    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    partner = my_id ^ pow2(step);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_addr         (rport, &r_result[start_recv]);
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner,
		rport,
		0,
		0,
		&r_result[start_send],
		data_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }
  show(my_id, 0, v_size, r_result);
}

/******************************************************************************
 *                                                                            *
 *  Function : CMMD_PRS0()  
 *                                                                            *
 *     - Scalable vector PRS on PS_i & RS_i                                   *
 *                                                                            *
 ******************************************************************************/
int CMMD_PRS0(size, p_result, r_result)
     int size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{

  CMMD_scan_v(p_result, p_result, CMMD_combiner_add,
              CMMD_upward, CMMD_none, 0, CMMD_exclusive,
              4, size);

  CMMD_reduce_v(r_result, r_result, CMMD_combiner_add, 4, size);

}

/******************************************************************************
 *                                                                            *
 *  Function : PRS()                                                          *
 *                                                                            *
 *     - Scalable vector PRS on PS_i & RS_i                                   *
 *                                                                            *
 ******************************************************************************/
int PRS_dim0(size, p_result, r_result)
     int size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{

  int out_mesg[Max], in_mesg[Max]; 

  int partner;
  int i, j;

  if ( size < 1 )
    return 0;

  for (j=0; j<size; ++j) {
    out_mesg[j] = p_result[j];  /* same as r_result */
    p_result[j] = 0;            /* exclusive prefix */
  }

  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  CMAML_set_rport_total_nbytes (rport, msg_size*size);

  CMMD_sync_with_nodes();

  for (i=0; i<Ilog2(nproc); ++i) {

    partner = my_id ^ pow2(i);

    prs_dim_ack = false;

    CMAML_set_rport_byte_counter (rport, msg_size*size);

    CMMD_sync_with_nodes();

    CMAML_scopy(partner,
		rport,
		0,
		0,
		out_mesg,
		msg_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    for (j=0; j<size; ++j)
      out_mesg[j] += in_mesg[j];

    if (partner < my_id)
      for (j=0; j<size; ++j)
	p_result[j] += in_mesg[j];
  }

  for (j=0; j<size; ++j)
    r_result[j] = out_mesg[j];

  CMAML_enable_interrupts();

  CMMD_sync_with_nodes();

  return 0;
}



/*******************************************************************************
 *                                                                             * 
 *                              MAIN  PROGRAM                                  *
 *                                                                             *
 *******************************************************************************/
  
main(argc, argv)
     int argc;
     char *argv[];
{
  FILE   *fd; /* input vector file name */

  int    i, j, k;
  int    local_size; /* Local vector size */
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  double time, time0, time1, time2;
  int p_result[Max], r_result[Max];
  int input_v[Max];
  double cmmd, prs;

  if (argc < 2) {
    printf("**************************************************\n");
    printf(" Usage: (1) the number of processor(s) ( 1 - 32 ) \n");
    printf(" Usage: (2) vector size \n");
    printf("**************************************************\n");
    exit(1);
  }

  /*-------------------------  Initialize  -----------------------------*/

  CMMD_disable_safety();
  CMMD_fset_io_mode(stdin, CMMD_sync_bc);
  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  CMMD_fset_io_mode(stderr, CMMD_independent);

  my_id = CMMD_self_address(); 

  if (i != nproc) 
    CMMD_error("\nERROR: CMMD Partition failed!\n");

  /** For Active Message Passing **/

  if ( CMAML_allocate_this_rport(rport) <= 0 )
    CMMD_error("ERROR : Failed to allocate receive rport %d \n", rport);

  /********************************/

  nproc  = atoi(argv[1]);
  v_size = atoi(argv[2]);

  i = CMMD_reset_partition_size(nproc);

/* Initialize the vector in each processor */

  for ( i = 0 ; i < v_size ; i++ ) {
    input_v[i] = 1;
  }

  /** For Active Message Passing **/
  CMAML_disable_interrupts();
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_node_timer_clear(0);
  CMMD_node_timer_clear(1);
  CMMD_node_timer_clear(2);

  CMMD_node_timer_start(2);

  CMMD_sync_with_nodes();
  CMMD_node_timer_start(0);
  ex_prefix(v_size, input_v, p_result);
  CMMD_node_timer_stop(0);

  CMMD_sync_with_nodes();
  CMMD_node_timer_start(1);
  reduction(v_size, input_v, r_result);
  CMMD_node_timer_stop(1);

  CMMD_node_timer_stop(2);

  time0 = CMMD_node_timer_busy(0) * 1000.;  /* milisec */
  time1 = CMMD_node_timer_busy(1) * 1000.;  /* milisec */
  time2 = CMMD_node_timer_busy(2) * 1000.;  /* milisec */

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n*********************************************************\n");
  printf("\n  Nproc       = %d", nproc);
  printf("\n  Vector size = %d\n", v_size);
  printf("\n  CMMD Vector exclusive prefix    = %5.3f msec", time0);
  printf("\n  CMMD Vector exclusive reduction = %5.3f msec", time1);
  printf("\n  CMMD Vector exclusive total     = %5.3f msec", time2);
  printf("\n\n*********************************************************\n\n");

  CMMD_sync_with_nodes();
  CMMD_node_timer_clear(0);
  CMMD_node_timer_clear(1);

  CMMD_node_timer_start(0);
  PRS_dim0(v_size, p_result, r_result);
  CMMD_node_timer_stop(0);

  CMMD_node_timer_start(1);
  CMMD_PRS0(v_size, p_result, r_result);
  CMMD_node_timer_stop(1);


  prs = CMMD_node_timer_busy(0) * 1000.;  /* milisec */
  cmmd= CMMD_node_timer_busy(1) * 1000.;  /* milisec */


  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n*******************************************\n");
  printf("\n  Nproc       = %d", nproc);
  printf("\n  Vector size = %d\n", v_size);
  printf("\n  CMMD Vector Scan/Reduction = %5.3f msec", cmmd);
  printf("\n  PRS on dim0                = %5.3f msec \n", prs);
  printf("\n*******************************************\n\n");
  CMAML_enable_interrupts();

}

